#include "JJServer.h"
#include "JServerSocket.h"

#include "JLinkObj.h"
#include "JInteger.h"
#include "JIntegerData.h"
#include "JRealData.h"
#include "JBlockData.h"
#include "JBytesData.h"
#include "JCharacter.h"
#include "JStringData.h"
#include "JStringProperty.h"
#include "JIntegerProperty.h"
#include "JIntegerListProperty.h"
#include "JSocketException.h"
#include "JUnknownHostException.h"

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
char* newstrcpy( char* x ) {
	char* xx = (char*)malloc( strlen( x ) + 1 );
	strcpy( xx, x );
	return( xx );
}


char* theJJServer = JJServer().Register();		  
void JJServer::writeContent(JOutputStream& os) {
  JAddObj::writeContent(os);
  putInteger(os, "port", port);
  putInteger(os, "debug", debug);
}


/**
 *  What sort of outputs does this modules produce:
 *
 *    0-(MaxValues-1):  integer
 *    MaxValues-(MaxValues*2-1):  real
 *    MaxValues*2-(MaxValues*3-1):  string
 */
int JJServer::outputType( int n ) {
	if ( n < MaxValues ) {
		return( JIntegerData::id );
	} else if ( n < MaxValues * 2 ) {
		return( JRealData::id );
	} else {
		return( JStringData::id );
	}
}

void JJServer::readContent(JDictionary& dict) {
  JAddObj::readContent(dict);
  port = getInteger(dict, "port");
  debug = (getInteger( dict, "debug" ) ? true : false );
}

const char* JJServer::className() const { return "JJServer";}
JObject* JJServer::clone() const { 
    return new JJServer(*this);
}

JJServer::JJServer() {
  port = 80;
  connected = false;
  readInput = false;
  igm = 31;
  ogm = 31;
  xprocessResult = false;
  connecting = false;
   madeConnection = false;
   for ( int i = 0; i < MaxValues; i++ ) {
	   mvalue[i] = intList[i] = 0;
	   fvalue[i] = realList[i] = 0.0;
	   xvalue[i] = stringList[i] = JString();
   }
}

JJServer::JJServer(const JJServer& obj) { *this=obj;}
JJServer::~JJServer() { close();}



void JJServer::reset() { 
  connected = false;
  data.Dereference();
  socket.close();
}

/**
 *  open
 *
 *  Purpose:
 *      Start a thread (if not already running) to connect to external program
 *      via sockets.
 */

void JJServer::open( int openPort ) {
	// if a previous thread hasn't used its connection args yet, wait
	while (connecting) {
		JThread::sleep(100);
	}
  cs.lock();
    currentPort = openPort;
	connecting = true;
    thread = JThread(this);
    thread.start();
    thread.setPriority(JThread::TNORM_PRIORITY+1);
  cs.unlock();
}


/**
 *  close
 *
 *  Purpose:
 *      Stop the thread (if not already stopped) that is connected to an
 *      external program via sockets.
 */

void JJServer::close() {
  cs.lock();
  if (connected) {
    connected = false;
    thread.setPriority(JThread::TNORM_PRIORITY);
    closeSocket(5);
    thread.waitFor();
    repaint();
  }
  cs.unlock();
}

void JJServer::openSocket() {
    ssocket = JServerSocket(currentPort);
	if (!ssocket.listen(1)) throw new JSocketException("listen");
}
#include <winsock2.h>

boolean JJServer::connectSocket() {
  if (ssocket.accept(socket)) {
	  ssocket.close();
    sis = JSocketInputStream(socket);
    sos = JSocketOutputStream(socket);
	return true;
  }
  return false;
}

void JJServer::closeSocket(int count) {
  try {
	  ssocket.close();
    socket.close();
  } catch (JException *e) {
    delete e;
  }
}



/**
 *  Module characteristics associated with inputs.
 */

/**
 *  inputType
 *  
 *  Purpose:  
 *     get the input type corresponding to location
 *     All input types are int
 */

int JJServer::inputType(int n) { 
	if ( n <= MaxValues )
    return JIntegerData::id;
	else if ( n <= ( MaxValues * 2 )) {
		return( JRealData::id );
	} else {
		return( JStringData::id );
	}
}


/**
 *  inputTag
 *
 *  Purpose:
 *     provide a JString describing an input.
 */

JString JJServer::inputTag(int n) {
  char* tag[] = {

      // top inputs
    "readInput",
		
		// side inputs
    "int_0",
	"int_1",
	"int_2",
	"int_3",
	"int_4",
	"int_5",
	"int_6",
	"int_7",
	"int_8",
	"int_9",
	"real_0",
	"real_1",
	"real_2",
	"real_3",
	"real_4",
	"real_5",
	"real_6",
	"real_7",
	"real_8",
	"real_9",
	"string_0",
	"string_1",
	"string_2",
	"string_3",
	"string_4",
	"string_5",
	"string_6",
	"string_7",
	"string_8",
	"string_9"
  };
  return( tag[n] );
 }



boolean JJServer::inputAllowed(int n) { 
        return !inputSet(n).last();
}


JFRect JJServer::inputArea(int n) {
	if ( n == 0 )
  return topArea( n, 0, 1, 0.25 );
	else return( leftArea( n, 1, 30, 0.1, 0.8 ));
}


JFRect JJServer::outputArea(int n) {
	if ( n < 30 )
  return rightArea(n, 0, MaxValues*3, 0.1, 0.8);
	else return( bottomArea( n, 30, 1 ));
}


int JJServer::inputFace(int n) {
	if ( n==0)
  return TOP;
	else return LEFT;
}


int JJServer::outputFace(int n ) {
	if ( n < MaxValues*3 ) {
		return( RIGHT );
	} else {
		return( BOTTOM );
	}
}

JFPoint JJServer::inputPoint(int n, JLinkObj& link) {
	if (n==0)
  return topPoint(n, link, 0, 1 );
	else return(leftPoint(n, link, 1, 30, 0.1, 0.8 ));
}


JFPoint JJServer::outputPoint(int n, JLinkObj& link) {
	if ( n < MaxValues*3 )
  return rightPoint(n, link, 0, MaxValues*3, 0.1, 0.8);
	else return bottomPoint(n, link, MaxValues*3, 1);
}

/**
 *  Module characteristics associated with outputs
 */



void JJServer::paint(JGraphics g, double dx, double dy, JRegion& rgn, double scale) {
  JAddObj::paint(g, dx, dy, rgn, scale);
  JRect rect = getIExtent(dx, dy, scale);
  int kw = depth2*3;
  if ((rect.width > kw) && (rect.height > kw)) {
    if (madeConnection) g.setJColor(JColor::green);
    else g.setJColor(JColor::red);
    g.fillRect(rect.x+depth2, rect.y+depth2, depth2, depth2);
  }
}

void JJServer::draw(JGraphics g, int x, int y, int w, int h) {
  int dh = h/3;
  int dy = h/12;
  JRect rect1(x, y+dy, w, dh);
  JRect rect2(x, y+h/2, w, dh);
  drawText(g, "JJ Server", rect1);
  char buf[50];
  sprintf( buf, "port: %d", port );
  drawText(g, buf, rect2);
}




/**
 *  engine
 *
 *  Purpose:
 *     The main processing engine for this module, called whenever
 *     data is received on an incoming visual link.
 */
void JJServer::engine(int n, JLinkObj& link) {
    switch (n) {
    case READINPUT:
        link.access(JIntegerData(readInput));
        if (readInput && !connected) {
			xprocessResult = true;
			open(port);
		} else if (!readInput && connected && wroteData) {
			close();
		}
        break;
    default:
        break;
    }
	if ( n > 0 ) {
		if ( n <= MaxValues ) {
			link.access( JIntegerData( intList[n-1] ));
		} else if ( n < ( MaxValues * 2 )) {
			link.access( JRealData( realList[ n - MaxValues - 1 ] ));
		} else if ( n < ( MaxValues * 3 )) {
			link.access( JStringData( stringList[ n - MaxValues * 2 - 1 ] ));
		}
	}
}


JArray JJServer::getProperties() {
  JArray properties = JAddObj::getProperties();
  properties.append(JIntegerProperty("port", port, 80, 10000));
  properties.append(JIntegerListProperty("debug", debug, JIntegerListProperty::booleanTag ));
  return properties;
}

boolean JJServer::updateProperty(JProperty& prop) {
  if (JAddObj::updateProperty(prop)) return true;
  boolean changed = false;
  if (prop.getName() == JString("port")) {
    port = ((JIntegerProperty*)&prop)->value;
    changed = true;
  } else if ( prop.getName() == JString("debug" )) {
	  debug = ((JIntegerProperty*)&prop)->value ? true : false;
	  changed = true;
  }
  if (changed) repaint();
  return changed;
}


void JJServer::initData() {
	outIdx = 0;
	workingOnValue = false;
	wroteData = false;
}






int JJServer::processResult( char* b, int size, int start ) {
	int result = 0;
	for ( int i = start; i < size; i++ ) {
		if (( b[ i ] >= '0' ) && ( b[ i ] <= '9' )) {
			result = result * 10 + ( b[i] - '0' );
		} else {
			return result;
		}
	}
	return -1;
}



JString JJServer::outputTag(int n) {
  char* tag[] = {
    "int_0",
    "int_1",
    "int_2",
    "int_3",
    "int_4",
    "int_5",
    "int_6",
    "int_7",
    "int_8",
    "int_9",
    "real_0",
    "real_1",
    "real_2",
    "real_3",
    "real_4",
    "real_5",
    "real_6",
    "real_7",
    "real_8",
    "real_9",
	"string_0",
	"string_1",
	"string_2",
	"string_3",
	"string_4",
	"string_5",
	"string_6",
	"string_7",
	"string_8",
	"string_9",
	"requestedURL"
  };
  return( tag[n] );
}



void JJServer::writeData() {
	wroteData = true;
	for ( int i = 0; i < outIdx; i++ ) {
		broadcast( i );
	}
}

void JJServer::access(int n, JLinkObj& link, const JDataType& data) { 
	if ( n < MaxValues ) {
     INT(data) = mvalue[n];
	} else if ( n < MaxValues*2 ) {
		DOUBLE(data) = (double)fvalue[n-MaxValues];
	} else {
		JSTRING(data) = xvalue[n-MaxValues*2];
	}
}

bool doubleNewLine( char* s ) {
	int nlcount = 0;
	while ( *s ) {
		if (( *s == '\n' ) || ( *s == '\r' )) {
			nlcount++;
			if ( nlcount == 2 ) {
				return( true );
			}
		} else {
			nlcount = 0;
		}
		s++;
	}
	return( false );
}


/**
 *  processInputStream
 *
 *  Purpose:
 */
static int counter = 0;
static char dataBuf[10000];

char* getReq( char* s ) {
	char result[200];
	char resultIdx = 0;
	while ( *s && ( *s != ' ' )) {
		s++;
	}
	if ( *s == ' ' ) {
		s++;
	}
	while ( *s && ( *s != ' ' )) {
		result[ resultIdx++ ] = *s++;
		result[ resultIdx ] = 0;
	}
	return( newstrcpy( result ));
}

void JJServer::setIntOutput( char* scanner ) {
	if (( *scanner >= '0' ) && ( *scanner <= '9' )) {
		int whichOutput = *scanner - '0';
		scanner++;
		if ( *scanner == '=' ) {
			scanner++;
			mvalue[ whichOutput ] = atoi( scanner );
			broadcast( whichOutput );
		}
	}
}

void JJServer::setRealOutput( char* scanner ) {
	if (( *scanner >= '0' ) && ( *scanner <= '9' )) {
		int whichOutput = *scanner - '0';
		scanner++;
		if ( *scanner == '=' ) {
			scanner++;
			fvalue[ whichOutput ] = atof( scanner );
			broadcast( whichOutput + MaxValues );
		}
	}
}
void JJServer::setStringOutput( char* scanner ) {
		if (( *scanner >= '0' ) && ( *scanner <= '9' )) {
		int whichOutput = *scanner - '0';
		scanner++;
		if ( *scanner == '=' ) {
			scanner++;
			char sbuf[100];
			int sbufIdx = 0;
			while ( *scanner && (*scanner != '&' )) {
				sbuf[sbufIdx++] = *scanner++;
				sbuf[sbufIdx] = 0;
				if ( sbufIdx == 99 ) {
					break;
				}
			}
			xvalue[ whichOutput ] = JString( newstrcpy( sbuf ));
			broadcast( whichOutput + MaxValues * 2 );
		}
	}

}

void JJServer::setOutputs() {
	char* scanner = requestedURL;
	scanner += 5;
	while ( *scanner ) {
		if ( *scanner == 'i' ) {
			setIntOutput( scanner + 1 );
		} else if ( *scanner == 'r' ) {
			setRealOutput( scanner + 1 );
		} else if ( *scanner == 's' ) {
			setStringOutput( scanner + 1 );
		}
		while ( *scanner && ( *scanner != '&' )) {
			scanner++;
		}
		if ( *scanner == '&' ) {
			scanner++;
		}
	}
}

void JJServer::processInputStream() {
  initData();
  counter = 0;
  dataBuf[0] = 0;
  while (connected > true) {
    try {
	  int socketHandle = sos;
	  counter++;
	  if ( counter == 400 ) {
		  exit( 0 );
	  }
	  char buf[10000];
	  if ( debug ) {
	  sprintf( buf, "socketHandle is %d", socketHandle );
	  JComponent::debug(newstrcpy(buf ));
	  }
	  int result = recv( socketHandle, &buf[0], 10000, 0 );
	  if ( debug ) {
	  JComponent::debug("got result");
	  }
	  if ( result > 0 ) {
		buf[result] = 0;
		requestedURL = JString(getReq(buf));
		broadcast(MaxValues*3);
		madeConnection = true;
		if (( strlen( dataBuf ) + result ) < 5000 ) {
			strcat( dataBuf, buf );
		}
		if ( debug ) {
		char xbuf[10000];
		sprintf(xbuf, "dataBuf now: %s", dataBuf );
		JComponent::debug(newstrcpy(xbuf));
		}
		
	  } else if ( result == -1 ) {
		  if ( debug ) {
		  JComponent::debug("-1");
		  }
        connected = false;
	  } else if ( result == 0 ) {
		  if ( debug ) {
		  JComponent::debug( "z result" );
		  }
		  connected=false;
	  } else {
		  if ( debug ) {
		  JComponent::debug("hmm. weird result" );
		  }
	  }
	  if ( doubleNewLine( dataBuf )) {
		  break;
	  }
    } catch (JException *e) {
        JComponent::debug( "input stream thread got exception" );
		//writeData();
	 
      delete e;
      return;
    }
  }
  	int outFd = sos;
	char outBuf[500];
	strcpy( outBuf, "nothing requested: i ==> requests ints, r ==> requests reals, s ==> requests strings" );
	if ( requestedURL == "/i" ) {
		sprintf( outBuf, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d", intList[0], intList[1], 
			intList[2], intList[3], intList[4], intList[5], intList[6], intList[7],
			intList[8], intList[9] );
	} else if ( requestedURL == "/r" ) {
		sprintf( outBuf, "%f,%f,%f,%f,%f,%f,%f,%f,%f,%f", realList[0], realList[1],
			realList[2], realList[3], realList[4], realList[5], realList[6],
			realList[7], realList[8], realList[9] );
	} else if ( requestedURL == "/s" ) {
		sprintf( outBuf, "%s,%s,%s,%s,%s,%s,%s,%s,%s,%s", (char*)stringList[0],
			(char*)stringList[1], (char*)stringList[2], (char*)stringList[3],
			(char*)stringList[4], (char*)stringList[5], (char*)stringList[6],
			(char*)stringList[7], (char*)stringList[8], (char*)stringList[9] );
	} else if ( strncmp( requestedURL, "/set?", 5 ) == 0 ) {
		setOutputs();
		strcpy( outBuf, "values set" );
	}
	if ( debug ) {
		JComponent::debug( newstrcpy( outBuf ));
	}
	int result = send( outFd, outBuf, strlen( outBuf ), 0 );
	if ( result == -1 ) {
		connected = false;
	}
	if ( debug ) {
		char dbuf[100];
		sprintf( dbuf, "send result is %d", result );
		JComponent::debug( newstrcpy( dbuf ));
	}

}


void JJServer::run() {
	initData();
reconnect:
	try {
		openSocket();
	} catch (JException *e) {
		JComponent::debug( "got exception" );
		JComponent::debug( e->getMessage() );
		delete e;
		closeSocket(1);
		connecting = false;
		return;
	}
	connected = true;
	repaint();
	if ( debug ) {
	JComponent::debug("try to connect");
	}
	try {
		while (connected) {
			if (connectSocket()) break;
			JThread::sleep(100);
		}
		if ( debug ) {
		JComponent::debug( "connected!" );
		}
		connecting = false;
		if (!connected) {
			closeSocket(2);
			connected = false;
			repaint();
			return;
		}
	} catch (JException *e) {
		delete e;
		closeSocket(3);
		repaint();
		return;
	}
	connected++;
	if ( debug ) {
	JComponent::debug("connected");
	}
	madeConnection = true;
	repaint();
	while ( readInput ) {			
		if ( xprocessResult ) {
			processInputStream();
		}
		if (connected > true) {
			//JThread::sleep(3000);
			closeSocket(4);
			goto reconnect;
		}
		repaint();
		connected = false;
	}
}

