#if !defined( _JJServer_h )
#define _JJServer_h

#include "JAddObj.h"
#include "JSocket.h"
#include "JServerSocket.h"
#include "JSocketInputStream.h"
#include "JSocketOutputStream.h"
#include "JThread.h"
#include "JCriticalSection.h"

#define MaxValues 10

class JJServer : public JAddObj {
  protected:
    virtual void writeContent(class JOutputStream& os);
    virtual void readContent(class JDictionary& dict);
  public:
    enum {
      READINPUT, INT_0, INT_1, INT_2, INT_3, INT_4, INT_5, INT_6, INT_7, INT_8, INT_9,
	REAL_0, REAL_1, REAL_2, REAL_3, REAL_4, REAL_5, REAL_6, REAL_7, REAL_8, REAL_9,
	STRING_0, STRING_1, STRING_2, STRING_3, STRING_4, STRING_5, STRING_6, STRING_7, STRING_8, STRING_9 };

    virtual const char* className() const;
    virtual JObject* clone() const;
    JJServer();
    JJServer(const JJServer& obj);
    virtual ~JJServer();
    virtual void reset();
    virtual void open( int openPort );
    virtual void close();
    virtual void openSocket();
    virtual boolean connectSocket();
    virtual void closeSocket(int n);
    virtual int inputType(int n);
    virtual boolean inputAllowed(int n);
    virtual JString inputTag(int n);
    virtual JString outputTag(int n);
    virtual JFRect inputArea(int n);
    virtual JFRect outputArea(int n);
	virtual int outputType(int n);
    virtual int inputFace(int n);
	virtual int outputFace(int n);
    virtual JFPoint inputPoint(int n, JLinkObj& link);
    virtual JFPoint outputPoint(int n, JLinkObj& link);
    virtual void paint(JGraphics g, double dx, double dy, JRegion& rgn, double scale);
    virtual void draw(JGraphics g, int x, int y, int w, int h);
    virtual void engine(int n, JLinkObj& link);
    virtual void run();
    virtual JArray getProperties();
    virtual boolean updateProperty(JProperty& prop);
    virtual void access(int n, JLinkObj& link, const JDataType& data);
    void processInputStream();
  protected:
	  void initData();
	  int processResult( char* b, int size, int start );
	int tempVal;

	void writeData();
	void setOutputs();  // set output values based on requestedURL
	void setIntOutput( char* scanner );
	void setRealOutput( char* scanner );
	void setStringOutput( char* scanner );
	bool madeConnection;
    JSocket socket;
	JServerSocket ssocket;
    JSocketInputStream sis;
    JSocketOutputStream sos;

    boolean connected;
	boolean wroteData;
    JThread thread;
    JArray data;
    JCriticalSection cs;
	int outIdx;
	int workingOnValue;

    // controls
    int readInput;

	// inputs
	int intList[MaxValues];
	double realList[MaxValues];
	JString stringList[MaxValues];

    // properties
    int port;
	bool debug;
	bool synchOutputsWithInputs;


	// since we can open one or other, need to track current
	boolean connecting;   // delays thread until other requests are finished...
	int currentPort;
	boolean xprocessResult;

    // outputs
    int mvalue[MaxValues];
	double fvalue[MaxValues];
	JString xvalue[MaxValues];
	JString requestedURL;
};

#endif
